import path from 'path';
import fs from 'fs-extra';
import JSZip from 'jszip';
import { createFilter } from '@rollup/pluginutils';
import { emit } from 'tsickle';
import { readConfigFile, parseJsonConfigFileContent, sys, createCompilerHost, createProgram } from 'typescript';

const tsToJs = path => path.replace(/\.tsx?$/i, ".js");
const humanlizePath = filepath => path.relative(process.cwd(), filepath);

var index = ((options = {}) => {
  var _options$include, _options$tsconfig;

  const isIncluded = createFilter((_options$include = options.include) !== null && _options$include !== void 0 ? _options$include : /\.tsx?$/i, options.exclude);
  const tsconfig = (_options$tsconfig = options.tsconfig) !== null && _options$tsconfig !== void 0 ? _options$tsconfig : "tsconfig.json";
  const zip = new JSZip();
  const emitted = new Map();
  let compilerOptions;
  let compilerHost;
  let tsickleHost;
  const plugin = {
    name: "tsickle",

    buildStart() {
      const loaded = readConfigFile(tsconfig, file => fs.readFileSync(file, "utf8"));
      const parsed = parseJsonConfigFileContent(loaded.config, sys, ".", {}, tsconfig);
      compilerOptions = parsed.options;
      if (compilerOptions.paths) for (const [from, to] of Object.entries(compilerOptions.paths)) compilerOptions.paths[from] = to.map(t => path.resolve(t).replace(/\\/g, "/"));
      compilerHost = createCompilerHost(compilerOptions);
      tsickleHost = {
        shouldSkipTsickleProcessing: () => false,
        shouldIgnoreWarningsForPath: () => false,
        pathToModuleName: name => name,
        fileNameToModuleId: name => name,
        options: compilerOptions,
        moduleResolutionHost: compilerHost,
        googmodule: false
      };
    },

    transform(_, id) {
      if (!isIncluded(id)) return;
      options.debug && console.log(`TSICKLE - INCLUDED (${humanlizePath(id)})`);
      const _id = id.split("?")[0];

      if (_id !== id && /\.vue$/i.test(_id)) {
        options.debug && console.log(`TSICKLE - VUE SFC (${humanlizePath(id)})`);
        throw new Error("TS inside Vue SFC is not supported for now");
      }

      const sourceFileName = id.replace(/\\/g, "/");
      const program = createProgram([sourceFileName], compilerOptions, compilerHost);
      emit(program, tsickleHost, (file, code) => emitted.set(file, code));
      const sourceFileAsJs = tsToJs(sourceFileName);

      for (const [file, source] of emitted) {
        if (!sourceFileAsJs.includes(file)) continue;
        options.debug && zip.file(`individual/${path.basename(file)}`, source);
        const sourceRel = humanlizePath(sourceFileName);
        const resRel = path.basename(sourceFileAsJs);
        options.debug && console.log(`TSICKLE - PROCESSED (${sourceRel} to ${resRel})`);
        return {
          code: source
        };
      }

      options.debug && console.log(`TSICKLE - MISSED (${humanlizePath(id)} in ${JSON.stringify(emitted)})`);
      return null;
    },

    renderChunk(code, chunk) {
      options.debug && zip.file(chunk.fileName, code);
      return null;
    },

    async generateBundle() {
      options.debug && this.emitFile({
        type: "asset",
        name: "tsickle.zip",
        source: await zip.generateAsync({
          type: "uint8array"
        })
      });
    }

  };
  return plugin;
});

export default index;
